# TODO: Bepaal mbv bestand 682 welke interacties, contra-indicaties, dubbelmedicaties, leeftijd als CI 
# (op basis van thes.17) en meldingen van bijzondere kenmerken niet meer hoeven te worden uitgevoerd.

from constants import ATTRIBUUT_GETAL, FUNCTIE_CI_AARD

def aanwezigheid_parameter_waardelijst(naald, hooiberg):
    if naald in hooiberg:
        return 100
    return 0


attribuut_functies = {
    4: aanwezigheid_parameter_waardelijst,
}
# attribuut_functies[4](3, (1, 2))

operators = {
    '=': (lambda x, y: x == y)
}
# operators['='](True, True)


def trav_actie(actie, debug=False):
    
    # TODO: moet er nog iets met 694/bouwsteen/vervolg actie gebeuren?
    assert len(actie.bouwstenen) == 0, 'Not implemented'
    
    if actie.mfbajn == 'J':
        desc = '\n'.join(actie.tekst)
        if debug:
            print('DESC:', desc)
        return True
    
    else:
        if debug:
            print('Geen actie:', actie.mfbaoms)
        return False


def trav_prot(flow, sample_params, score_teller, doorlopen_pad, debug=False):
    
    # Vragen zijn onderdeel van een knooppunt in het protocol (zie bst 691).
    # Het antwoord op de vraag (ja of nee) bepaalt het vervolgknooppunt binnen het protocol.
    # Vragen bevatten in principe een functies, attributen en zo nodig parameters en/of waardenlijsten.
    # Vragen die geen functie, attribuut, parameter of waardenlijst hebben, bevatten een protocolattribuut.
    
    if debug:
        print('protocol flow:', flow.mfbpnr)
        print('knoop nummer:', flow.mfbknr)
        
    vraag = flow.vraag
    
    if debug:
        print('Vraagnummer: %s' % vraag.mfbvnr)
        print('Vraag omschrijving:', vraag.mfbvoms)
        print('Vraag operator:', vraag.mfbvoper)

        # TODO: deal better with decimal values
        print('Vraag waarde: %d' % vraag.mfbvw)
                        
    # We don't have code to deal with protocol attributes
    # (Geen eerder berekend resultaat gebruiken.)
    assert vraag.mfbfuwo == 0, 'Protocol attribuut ophalen: %s' % vraag.mfbfuwo
    
    # Validate that questions have a function (we ignored attributed above ...)
    assert vraag.mfbfunnr > 0, 'Geen functie gedefinieerd'
    
    if debug:
        print('Functie omschrijving: %s' % vraag.functie.mfbfuoms)
    
    # We can only deal with "Zoek naar een bepaalde CI-aard bij de patient"?!
    if vraag.mfbfunnr != FUNCTIE_CI_AARD:
        print('WARNING, non-FUNCTIE_CI_AARD function "%d:%s"' % (
            vraag.mfbfunnr, vraag.functie.mfbfuoms))
    
    # "Door MFBFUNS1 kunnen meerdere parameters aan een functie (bij een specifieke vraag) gekoppeld worden."
    assert len(vraag.parameters) == 1, 'Dont know how to handle more or less'
    parameter = vraag.parameters[0]
    if debug:
        print('Parameter:', parameter.mfbpanr, parameter.mfbpaoms)
        print('Sample parameters: %s' % sample_params)
        
    # "Door MFBFUNS2 kunnen meerdere waardenlijsten aan een functie (bij een specifieke vraag) gekoppeld worden."
    # Vooralsnog altijd leeg in onze data
    assert len(vraag.waardelijsten) == 0, "Not implemented"
   
    # "Door MFBFUNS3 kunnen meerdere attributen aan een functie (bij een specifieke vraag) gekoppeld worden."
    # Kunnen er potentieel meer zijn
    assert len(vraag.attributen) == 1, 'Dont know how to handle more or less'
    attribuut = vraag.attributen[0]
    if debug:
        print('Attribuut %d, %s:' % (attribuut.mfbatnr, attribuut.mfbatoms))
        print('Attribuut type: %s (%s)' % (attribuut.mfbattyp, 'GETAL' if attribuut.mfbattyp == ATTRIBUUT_GETAL else 'BOOLEAN'))

    # We don't have code to deal with protocol attributes
    assert vraag.vraag_functie_attribuut[0].mfbfuwt == 0, 'Protocol attribuut wegschrijven: %s' % vraag.vraag_functie_attribuut[0].mfbfuwt
    
    # Currently we only have one attribute function implemented
    assert attribuut.mfbatnr in [4], "Geen implementatie voor %d, %s" % (attribuut.mfbatnr, attribuut.mfbatoms)
    
    IVwaarde = attribuut_functies[attribuut.mfbatnr](parameter.mfbpanr, sample_params)   
    if debug:
        print('IVwaarde: %d' % IVwaarde)
        
    op_result = operators[vraag.mfbvoper.strip()](IVwaarde, vraag.mfbvw)
    if debug:
        print("Operator result:", op_result)
    
    if op_result:
        score_teller += vraag.mfbvstj
        doorlopen_pad.append(vraag.mfbvstjt)

        if flow.mfbpjk:
            trav_prot(flow.ja_flow, sample_params, score_teller, doorlopen_pad, debug)
        else:
            if debug:
                print('JA Actie: %s' % flow.mfbpja)
            return trav_actie(flow.ja_actie, debug)
    else:
        score_teller += vraag.mfbvstn
        doorlopen_pad.append(vraag.mfbvstnt)

        if flow.mfbpnk:
            trav_prot(flow.nee_flow, sample_params, score_teller, doorlopen_pad, debug)
        else:
            if debug:
                print('NEE Actie: %s' % flow.mfbpna)
            return trav_actie(flow.nee_actie, debug)



from parsing import get_bestand_html, extract_struct
from files import file_numbers
from definition import foreign_keys, proxies, aggregates, relationships
from constants import SKIP_FIELDS


def get_non_composite_key(table_name, name, debug=False):
    if table_name not in foreign_keys:
        return None

    # Find out if there are multiple foreign keys per target table
    for group in foreign_keys[table_name]:

        if name in group:
            target_table, member = group[name].split('.')

            if debug:
                print('Target table: %s, member: %s' % (target_table, member))

            if table_name == target_table:
                if debug:
                    print('%s is a self reference' % target_table)

                return group[name]

            elif len(group) > 1:

                if debug:
                    print('%s is part of a composite foreign key' % name)
                return None

            else:
                if debug:
                    print('%s is not part of a composite foreign key' % name)
                return group[name]

    return None


def get_target_name(table_name, name):
    if table_name not in foreign_keys:
        return None

    for group in foreign_keys[table_name]:
        if name in group:
            return group[name]

    return None


def create_class(file_no):

    code = ''

    table_name = 'bst_%s' % file_no
    class_name = table_name

    code += '\n\nclass %s(Base):\n' % class_name

    html = get_bestand_html(file_no)
    fields = extract_struct(html.body.find('table', attrs={'class': 'zindextable'}))

    title = html.body.find('h1', attrs={'class': 'documentFirstHeading'}).string

    code += '    """Title: %s"""\n' % title
    code += '    __tablename__ = "%s"\n' % table_name
    code += '\n'

    fks = []
    repr_fields = []
    for s in sorted(fields, key=lambda x: x['key']):

        name = s['name']

        if name in SKIP_FIELDS:
            continue

        if name != 'bstnum':
            repr_fields.append(name)

        kind = s['kind']
        assert kind in ['Integer', 'String']

        key = s['key']
        assert key is not None

        size = s['size']
        assert size > 0

        fk = get_non_composite_key(table_name, name)
        if not fk:
            target_name = get_target_name(table_name, name)
            if target_name:
                fks.append((name, target_name))

        code += '    %s = Column(%s%s%s%s)\n' % (
            name,
            kind,
            '(%s)' % size if kind == 'String' else '',
            ', ForeignKey("%s")' % fk if fk else '',
            ', primary_key=True' if key > 0 else '')

    code += '\n'
    code += '    def __repr__(self):\n'
    code += '        return "<%s(%%d)[%s]>" %% (self.bstnum, %s)\n' % (
        title.split(' ', 3)[-1],
        ', '.join(['%s=%%s' % x for x in repr_fields]),
        ', '.join(['self.' + x for x in repr_fields])
    )

    if table_name in relationships:
        code += '\n'
        for rel in relationships[table_name]:
            entry = relationships[table_name][rel]

            # We use square brackets to denote that the member should be a list
            if not isinstance(entry, str):
                assert len(entry) == 1
                uselist = True
                entry = entry[0]
            else:
                uselist = False

            def entry2x(entry):
                local, remote = entry.split('==')

                if '.' in local:
                    local_table, local_column = local.split('.')
                else:
                    local_table, local_column = None, None

                if '.' in remote:
                    remote_table, remote_column = remote.split('.')
                else:
                    remote_table, remote_column = None, None

                return local, local_table, local_column, remote, remote_table, remote_column

            # Check the number of entries
            if ',' not in entry:
                # single entry

                if '==' not in entry:
                    # simple relationship
                    desc = entry
                else:
                    # complex relationship
                    _, lt, _, _, rt, rc = entry2x(entry)
                    desc = "'%s'" % rt

                    if uselist:
                        desc += ', uselist=True'

                    if lt == rt:
                        desc += ', remote_side=[%s]' % rc
                    desc += ", primaryjoin=%s" % entry

            else:
                # Multiple entries
                remote_columns = []
                remote_tables = []
                local_tables = []

                entries = [x.strip() for x in entry.split(',')]

                for one in entries:
                    local, lt, _, remote, rt, rc = entry2x(one)

                    if rc:
                        remote_columns.append(rc)
                    if rt:
                        remote_tables.append(rt)
                    if lt:
                        local_tables.append(lt)

                assert len(set(local_tables)) == 1
                assert len(set(remote_tables)) == 1

                desc = "'%s'" % remote_tables[0]

                if uselist:
                    desc += ', uselist=True'

                desc += ", primaryjoin='%s'" % 'and_(%s)' % ', '.join(entries)

                if local_tables[0] == remote_tables[0]:
                    desc += ', remote_side=[%s]' % ', '.join(remote_columns)

            # End of relationship construction
            code += '    %s = relationship(%s)\n' % (rel, desc)

    if table_name in aggregates:
        for new_entry in aggregates[table_name]:
            code += '\n'
            code += '    @hybrid_property\n'
            code += '    def %s(self):\n' % new_entry
            code += '        return [x.%s for x in %s]\n' % aggregates[table_name][new_entry]  # Tuple

    if table_name in proxies:
        code += '\n'
        for kw in proxies[table_name]:
            code += '    %s = association_proxy("%s", "%s")\n' % (
                         kw, proxies[table_name][kw][0], proxies[table_name][kw][1])

    if fks:
        lhs = []
        rhs = []

        for fk, target_name in fks:
            lhs.append(fk)
            rhs.append(target_name)

        code += '\n'
        code += '    __table_args__ = (\n'
        code += '        ForeignKeyConstraint(\n'
        code += '            %s,\n' % lhs
        code += '            %s,\n' % rhs
        code += '        ),\n'
        code += '    )\n'

    return code


all_code = '''\
#
# !!!! this file is auto-generated !!!
#

from sqlalchemy import Column, Integer, String, ForeignKey, ForeignKeyConstraint
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.ext.associationproxy import association_proxy

Base = declarative_base()
'''


for file_no in file_numbers:
    all_code += create_class(file_no)

print(all_code)



#
# !!!! this file is auto-generated !!!
#

from sqlalchemy import Column, Integer, String, ForeignKey, ForeignKeyConstraint
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.ext.associationproxy import association_proxy

Base = declarative_base()


class bst_685(Base):
    """Title: Bestandsbeschrijvingen: Bestand 685 MFB: Parameters"""
    __tablename__ = "bst_685"

    bstnum = Column(Integer)
    mutkod = Column(Integer)
    mfbvnopn = Column(Integer)
    mfbvnwyz = Column(Integer)
    mfbvnvv = Column(Integer)
    mfbpaoms = Column(String(80))
    mfbpavt = Column(Integer)
    thmfbp = Column(Integer)
    mfbpitnr = Column(Integer)
    thmodu = Column(Integer)
    txmodu = Column(Integer)
    mfbpanr = Column(Integer, primary_key=True)

    def __repr__(self):
        return "<MFB: Parameters(%d)[mutkod=%s, mfbvnopn=%s, mfbvnwyz=%s, mfbvnvv=%s, mfbpaoms=%s, mfbpavt=%s, thmfbp=%s, mfbpitnr=%s, thmodu=%s, txmodu=%s, mfbpanr=%s]>" % (self.bstnum, self.mutkod, self.mfbvnopn, self.mfbvnwyz, self.mfbvnvv, self.mfbpaoms, self.mfbpavt, self.thmfbp, self.mfbpitnr, self.thmodu, self.txmodu, self.mfbpanr)

    thesaurus = relationship("bst_902")

    __table_args__ = (
        ForeignKeyConstraint(
            ['thmfbp', 'mfbpitnr'],
            ['bst_902.tsnr', 'bst_902.tsitnr'],
        ),
    )


class bst_686(Base):
    """Title: Bestandsbeschrijvingen: Bestand 686 MFB: Attributen"""
    __tablename__ = "bst_686"

    bstnum = Column(Integer)
    mutkod = Column(Integer)
    mfbattyp = Column(Integer)
    mfbvnopn = Column(Integer)
    mfbvnwyz = Column(Integer)
    mfbvnvv = Column(Integer)
    mfbatoms = Column(String(80))
    thmfbp = Column(Integer)
    mfbpitnr = Column(Integer)
    thmodu = Column(Integer)
    txmodu = Column(Integer)
    mfbatnr = Column(Integer, primary_key=True)

    def __repr__(self):
        return "<MFB: Attributen(%d)[mutkod=%s, mfbattyp=%s, mfbvnopn=%s, mfbvnwyz=%s, mfbvnvv=%s, mfbatoms=%s, thmfbp=%s, mfbpitnr=%s, thmodu=%s, txmodu=%s, mfbatnr=%s]>" % (self.bstnum, self.mutkod, self.mfbattyp, self.mfbvnopn, self.mfbvnwyz, self.mfbvnvv, self.mfbatoms, self.thmfbp, self.mfbpitnr, self.thmodu, self.txmodu, self.mfbatnr)


class bst_689(Base):
    """Title: Bestandsbeschrijvingen: Bestand 689 MFB: Functie"""
    __tablename__ = "bst_689"

    bstnum = Column(Integer)
    mutkod = Column(Integer)
    mfbvnopn = Column(Integer)
    mfbvnwyz = Column(Integer)
    mfbvnvv = Column(Integer)
    mfbfuoms = Column(String(80))
    thmodu = Column(Integer)
    txmodu = Column(Integer)
    mfbfunnr = Column(Integer, primary_key=True)

    def __repr__(self):
        return "<MFB: Functie(%d)[mutkod=%s, mfbvnopn=%s, mfbvnwyz=%s, mfbvnvv=%s, mfbfuoms=%s, thmodu=%s, txmodu=%s, mfbfunnr=%s]>" % (self.bstnum, self.mutkod, self.mfbvnopn, self.mfbvnwyz, self.mfbvnvv, self.mfbfuoms, self.thmodu, self.txmodu, self.mfbfunnr)


class bst_690(Base):
    """Title: Bestandsbeschrijvingen: Bestand 690 MFB: Protocol"""
    __tablename__ = "bst_690"

    bstnum = Column(Integer)
    mutkod = Column(Integer)
    mfbpdvv = Column(Integer)
    mfbpoms = Column(String(80))
    mfbpwin = Column(String(1))
    thmfbb = Column(Integer)
    mfbbron = Column(Integer)
    mfbpwind = Column(Integer)
    mfbknr = Column(Integer)
    thmodu = Column(Integer)
    txmodu = Column(Integer)
    mfbpnr = Column(Integer, primary_key=True)
    mfbpnrv = Column(Integer, primary_key=True)

    def __repr__(self):
        return "<MFB: Protocol(%d)[mutkod=%s, mfbpdvv=%s, mfbpoms=%s, mfbpwin=%s, thmfbb=%s, mfbbron=%s, mfbpwind=%s, mfbknr=%s, thmodu=%s, txmodu=%s, mfbpnr=%s, mfbpnrv=%s]>" % (self.bstnum, self.mutkod, self.mfbpdvv, self.mfbpoms, self.mfbpwin, self.thmfbb, self.mfbbron, self.mfbpwind, self.mfbknr, self.thmodu, self.txmodu, self.mfbpnr, self.mfbpnrv)

    flow = relationship("bst_691")

    __table_args__ = (
        ForeignKeyConstraint(
            ['mfbknr', 'mfbpnr'],
            ['bst_691.mfbknr', 'bst_691.mfbpnr'],
        ),
    )


class bst_691(Base):
    """Title: Bestandsbeschrijvingen: Bestand 691 MFB: Protocol flow"""
    __tablename__ = "bst_691"

    bstnum = Column(Integer)
    mutkod = Column(Integer)
    mfbpjk = Column(Integer, ForeignKey("bst_691.mfbknr"))
    mfbpja = Column(Integer, ForeignKey("bst_693.mfbanr"))
    mfbpnk = Column(Integer, ForeignKey("bst_691.mfbknr"))
    mfbpna = Column(Integer, ForeignKey("bst_693.mfbanr"))
    mfbvnr = Column(Integer, ForeignKey("bst_692.mfbvnr"))
    mfbpnr = Column(Integer, primary_key=True)
    mfbpnrv = Column(Integer, primary_key=True)
    mfbknr = Column(Integer, primary_key=True)

    def __repr__(self):
        return "<MFB: Protocol flow(%d)[mutkod=%s, mfbpjk=%s, mfbpja=%s, mfbpnk=%s, mfbpna=%s, mfbvnr=%s, mfbpnr=%s, mfbpnrv=%s, mfbknr=%s]>" % (self.bstnum, self.mutkod, self.mfbpjk, self.mfbpja, self.mfbpnk, self.mfbpna, self.mfbvnr, self.mfbpnr, self.mfbpnrv, self.mfbknr)

    vraag = relationship("bst_692")
    ja_flow = relationship('bst_691', primaryjoin='and_(bst_691.mfbpjk==bst_691.mfbknr, bst_691.mfbpnr==bst_691.mfbpnr)', remote_side=[mfbknr, mfbpnr])
    nee_flow = relationship('bst_691', primaryjoin='and_(bst_691.mfbpnk==bst_691.mfbknr, bst_691.mfbpnr==bst_691.mfbpnr)', remote_side=[mfbknr, mfbpnr])
    ja_actie = relationship('bst_693', primaryjoin='bst_691.mfbpja==bst_693.mfbanr')
    nee_actie = relationship('bst_693', primaryjoin='bst_691.mfbpna==bst_693.mfbanr')


class bst_692(Base):
    """Title: Bestandsbeschrijvingen: Bestand 692 MFB: Vragen"""
    __tablename__ = "bst_692"

    bstnum = Column(Integer)
    mutkod = Column(Integer)
    mfbvoms = Column(String(80))
    mfbfuwo = Column(Integer)
    mfbfunnr = Column(Integer, ForeignKey("bst_689.mfbfunnr"))
    mfbvstj = Column(Integer)
    mfbvstjt = Column(String(80))
    mfbvstn = Column(Integer)
    mfbvstnt = Column(String(80))
    mfbvoper = Column(String(2))
    mfbvw = Column(Integer)
    mfbvnr = Column(Integer, primary_key=True)

    def __repr__(self):
        return "<MFB: Vragen(%d)[mutkod=%s, mfbvoms=%s, mfbfuwo=%s, mfbfunnr=%s, mfbvstj=%s, mfbvstjt=%s, mfbvstn=%s, mfbvstnt=%s, mfbvoper=%s, mfbvw=%s, mfbvnr=%s]>" % (self.bstnum, self.mutkod, self.mfbvoms, self.mfbfuwo, self.mfbfunnr, self.mfbvstj, self.mfbvstjt, self.mfbvstn, self.mfbvstnt, self.mfbvoper, self.mfbvw, self.mfbvnr)

    functie = relationship("bst_689")
    vraag_functie_parameter = relationship('bst_695')
    vraag_functie_waardelijst = relationship('bst_696')
    vraag_functie_attribuut = relationship('bst_697')

    parameters = association_proxy("vraag_functie_parameter", "parameter")
    waardelijsten = association_proxy("vraag_functie_waardelijst", "waardelijst")
    attributen = association_proxy("vraag_functie_attribuut", "attribuut")


class bst_693(Base):
    """Title: Bestandsbeschrijvingen: Bestand 693 MFB: Actie"""
    __tablename__ = "bst_693"

    bstnum = Column(Integer)
    mutkod = Column(Integer)
    mfbaoms = Column(String(80))
    mfbajn = Column(String(1))
    mfbmon = Column(String(1))
    thmodu = Column(Integer)
    txmodu = Column(Integer, ForeignKey("bst_921.txmodu"))
    mfbanr = Column(Integer, primary_key=True)

    def __repr__(self):
        return "<MFB: Actie(%d)[mutkod=%s, mfbaoms=%s, mfbajn=%s, mfbmon=%s, thmodu=%s, txmodu=%s, mfbanr=%s]>" % (self.bstnum, self.mutkod, self.mfbaoms, self.mfbajn, self.mfbmon, self.thmodu, self.txmodu, self.mfbanr)

    teksten = relationship('bst_921', uselist=True, primaryjoin='and_(bst_693.mfbanr==bst_921.txkode, bst_693.txmodu==bst_921.txmodu, 240==bst_921.txtsrt)')
    bouwstenen = relationship("bst_694")

    @hybrid_property
    def tekst(self):
        return [x.txtext for x in self.teksten]


class bst_694(Base):
    """Title: Bestandsbeschrijvingen: Bestand 694 MFB: Actie - Parameter"""
    __tablename__ = "bst_694"

    bstnum = Column(Integer)
    mutkod = Column(Integer)
    mfbanr = Column(Integer, ForeignKey("bst_693.mfbanr"), primary_key=True)
    mfbnr = Column(Integer, primary_key=True)
    mfbaanst = Column(Integer, primary_key=True)

    def __repr__(self):
        return "<MFB: Actie - Parameter(%d)[mutkod=%s, mfbanr=%s, mfbnr=%s, mfbaanst=%s]>" % (self.bstnum, self.mutkod, self.mfbanr, self.mfbnr, self.mfbaanst)


class bst_695(Base):
    """Title: Bestandsbeschrijvingen: Bestand 695 MFB: Vraag - Functie - Parameter"""
    __tablename__ = "bst_695"

    bstnum = Column(Integer)
    mutkod = Column(Integer)
    mfbpanr = Column(Integer, ForeignKey("bst_685.mfbpanr"))
    mfbvnr = Column(Integer, ForeignKey("bst_692.mfbvnr"), primary_key=True)
    mfbfunnr = Column(Integer, ForeignKey("bst_689.mfbfunnr"), primary_key=True)
    mfbfuns1 = Column(Integer, primary_key=True)

    def __repr__(self):
        return "<MFB: Vraag - Functie - Parameter(%d)[mutkod=%s, mfbpanr=%s, mfbvnr=%s, mfbfunnr=%s, mfbfuns1=%s]>" % (self.bstnum, self.mutkod, self.mfbpanr, self.mfbvnr, self.mfbfunnr, self.mfbfuns1)

    vraag = relationship("bst_692")
    functie = relationship("bst_689")
    parameter = relationship("bst_685")


class bst_696(Base):
    """Title: Bestandsbeschrijvingen: Bestand 696 MFB: Vraag - Functie - Waardelijst"""
    __tablename__ = "bst_696"

    bstnum = Column(Integer)
    mutkod = Column(Integer)
    mfbwnr = Column(Integer, ForeignKey("bst_699.mfbwnr"))
    mfbvopew = Column(String(2))
    mfbvnr = Column(Integer, ForeignKey("bst_692.mfbvnr"), primary_key=True)
    mfbfunnr = Column(Integer, ForeignKey("bst_689.mfbfunnr"), primary_key=True)
    mfbfuns2 = Column(Integer, primary_key=True)

    def __repr__(self):
        return "<MFB: Vraag - Functie - Waardelijst(%d)[mutkod=%s, mfbwnr=%s, mfbvopew=%s, mfbvnr=%s, mfbfunnr=%s, mfbfuns2=%s]>" % (self.bstnum, self.mutkod, self.mfbwnr, self.mfbvopew, self.mfbvnr, self.mfbfunnr, self.mfbfuns2)

    vraag = relationship("bst_692")
    functie = relationship("bst_689")
    waardelijst = relationship("bst_699")


class bst_697(Base):
    """Title: Bestandsbeschrijvingen: Bestand 697 MFB: Vraag - Functie - Attribuut"""
    __tablename__ = "bst_697"

    bstnum = Column(Integer)
    mutkod = Column(Integer)
    mfbatnr = Column(Integer, ForeignKey("bst_686.mfbatnr"))
    mfbfuwt = Column(Integer)
    mfbvnr = Column(Integer, ForeignKey("bst_692.mfbvnr"), primary_key=True)
    mfbfunnr = Column(Integer, ForeignKey("bst_689.mfbfunnr"), primary_key=True)
    mfbfuns3 = Column(Integer, primary_key=True)

    def __repr__(self):
        return "<MFB: Vraag - Functie - Attribuut(%d)[mutkod=%s, mfbatnr=%s, mfbfuwt=%s, mfbvnr=%s, mfbfunnr=%s, mfbfuns3=%s]>" % (self.bstnum, self.mutkod, self.mfbatnr, self.mfbfuwt, self.mfbvnr, self.mfbfunnr, self.mfbfuns3)

    vraag = relationship("bst_692")
    functie = relationship("bst_689")
    attribuut = relationship("bst_686")


class bst_699(Base):
    """Title: Bestandsbeschrijvingen: Bestand 699 MFB: Waardelijsten meerdere niveaus"""
    __tablename__ = "bst_699"

    bstnum = Column(Integer)
    mutkod = Column(Integer)
    mfbwoms = Column(String(80))
    thsrtcde = Column(Integer)
    mfbwnr = Column(Integer, primary_key=True)
    srtcode = Column(Integer, primary_key=True)
    codenv = Column(String(10), primary_key=True)

    def __repr__(self):
        return "<MFB: Waardelijsten meerdere niveaus(%d)[mutkod=%s, mfbwoms=%s, thsrtcde=%s, mfbwnr=%s, srtcode=%s, codenv=%s]>" % (self.bstnum, self.mutkod, self.mfbwoms, self.thsrtcde, self.mfbwnr, self.srtcode, self.codenv)


class bst_902(Base):
    """Title: Bestandsbeschrijvingen: Bestand 902 Thesauri totaal"""
    __tablename__ = "bst_902"

    bstnum = Column(Integer)
    mutkod = Column(Integer)
    thitmk = Column(String(2))
    thnm4 = Column(String(4))
    thnm15 = Column(String(15))
    thnm25 = Column(String(25))
    thnm50 = Column(String(50))
    thakd1 = Column(String(1))
    thakd2 = Column(String(1))
    thakd3 = Column(String(1))
    thakd4 = Column(String(1))
    thakd5 = Column(String(1))
    thakd6 = Column(String(1))
    tsnr = Column(Integer, primary_key=True)
    tsitnr = Column(Integer, primary_key=True)

    def __repr__(self):
        return "<Thesauri totaal(%d)[mutkod=%s, thitmk=%s, thnm4=%s, thnm15=%s, thnm25=%s, thnm50=%s, thakd1=%s, thakd2=%s, thakd3=%s, thakd4=%s, thakd5=%s, thakd6=%s, tsnr=%s, tsitnr=%s]>" % (self.bstnum, self.mutkod, self.thitmk, self.thnm4, self.thnm15, self.thnm25, self.thnm50, self.thakd1, self.thakd2, self.thakd3, self.thakd4, self.thakd5, self.thakd6, self.tsnr, self.tsitnr)


class bst_921(Base):
    """Title: Bestandsbeschrijvingen: Bestand 921 Tekstblokken ASCII (vervangt 920)"""
    __tablename__ = "bst_921"

    bstnum = Column(Integer)
    mutkod = Column(Integer)
    thmodu = Column(Integer)
    thtsrt = Column(Integer)
    txtext = Column(String(132))
    txmodu = Column(Integer, primary_key=True)
    txtsrt = Column(Integer, primary_key=True)
    txkode = Column(Integer, primary_key=True)
    txblnr = Column(Integer, primary_key=True)
    txrgln = Column(Integer, primary_key=True)

    def __repr__(self):
        return "<Tekstblokken ASCII (vervangt 920)(%d)[mutkod=%s, thmodu=%s, thtsrt=%s, txtext=%s, txmodu=%s, txtsrt=%s, txkode=%s, txblnr=%s, txrgln=%s]>" % (self.bstnum, self.mutkod, self.thmodu, self.thtsrt, self.txtext, self.txmodu, self.txtsrt, self.txkode, self.txblnr, self.txrgln)




# We have data for these file numbers
file_numbers = {
#     '031': {}, # Handelsproducten
#     '051': {}, # Voorschrijfpr. geneesmiddel identific.
#     '601': {}, # Aanv.voorwaarden Med.Bew. en tekstblok
#     '655': {}, # Bewakingen: te controleren indicaties
#     '656': {}, # Bewakingen: Kenmerken
#     '681': {}, # MFB: Trigger geneesmiddelkeuze
#     '682': {}, # MFB: Vervanging medicatiebewaking
#     '684': {}, # MFB: Koppeling externe coderingen
    '685': {}, # MFB: Parameters
    '686': {}, # MFB: Attributen
    '689': {}, # MFB: Functie
    '690': {}, # MFB: Protocol
    '691': {}, # MFB: Protocol flow
    '692': {}, # MFB: Vragen
    '693': {}, # MFB: Actie
    '694': {}, # MFB: Actie - Parameter
    '695': {}, # MFB: Vraag - Functie - Parameter
    '696': {}, # MFB: Vraag - Functie - Waardelijst
    '697': {}, # MFB: Vraag - Functie - Attribuut
    '699': {}, # MFB: Waardelijsten meerdere niveaus
#     '711': {}, # Generieke producten
    '902': {}, # Thesauri totaal
#     '911': {}, # Header relatie bestand
#     '912': {}, # Relatie bestand
    '921': {}, # Tekstblokken ASCII
#     '922': {}, # Tekstblokken HTML
}



import importlib

def load_class(full_class_string):
    """
    dynamically load a class from a string
    credit: https://thomassileo.name/blog/2012/12/21/dynamically-load-python-modules-or-classes/
    """

    class_data = full_class_string.split(".")
    module_path = ".".join(class_data[:-1])
    class_str = class_data[-1]

    module = importlib.import_module(module_path)
    # Finally, we retrieve the Class
    return getattr(module, class_str)



# Database
database = 'foo.db'
# database = ':memory:'


BESTAND_BESCHRIJVINGEN_URL_PREFIX = 'https://www.z-index.nl/documentatie/bestandsbeschrijvingen/bestand?bestandsnaam='
BESCHRIJVINGEN_DIRECTORY = '/Users/mark/proj/pongs/zindex-exploration/beschrijvingen'
DATA_DIRECTORY = '/Users/mark/proj/pongs/zindex-exploration/data'



from constants import ZIEKENHUISTEKST

foreign_keys = {
    'bst_685': [
        {
            'mfbpitnr': 'bst_902.tsitnr',
            'thmfbp': 'bst_902.tsnr'
        }
    ],

    'bst_690': [
        {
            'mfbpnr': 'bst_691.mfbpnr',
            'mfbknr': 'bst_691.mfbknr'
        }
    ],

    'bst_691': [
        {
            'mfbvnr': 'bst_692.mfbvnr'
        },
        {
            'mfbpjk': 'bst_691.mfbknr'  # nullable=True ?
        },
        {
            'mfbpnk': 'bst_691.mfbknr'  # nullable=True ?
        },
        {
            'mfbpja': 'bst_693.mfbanr'
        },
        {
            'mfbpna': 'bst_693.mfbanr'
        }

    ],

    'bst_692': [
        {
            'mfbfunnr': 'bst_689.mfbfunnr'
        }
    ],

    'bst_693': [
        {
            'txmodu': 'bst_921.txmodu'
        }
    ],

    'bst_694': [
        {
            'mfbanr': 'bst_693.mfbanr'
        }
    ],

    'bst_695': [
        {
            'mfbvnr': 'bst_692.mfbvnr',
        },
        {
            'mfbfunnr': 'bst_689.mfbfunnr',
        },
        {
            'mfbpanr': 'bst_685.mfbpanr'
        }
    ],

    'bst_696': [
        {
            'mfbvnr': 'bst_692.mfbvnr',
        },
        {
            'mfbfunnr': 'bst_689.mfbfunnr',
        },
        {
            'mfbwnr': 'bst_699.mfbwnr'
        }
    ],

    'bst_697': [
        {
            'mfbvnr': 'bst_692.mfbvnr',
        },
        {
            'mfbfunnr': 'bst_689.mfbfunnr',
        },
        {
            'mfbatnr': 'bst_686.mfbatnr'
        }
    ],

}

relationships = {
    'bst_685': {
        'thesaurus': '"bst_902"'
    },

    'bst_690': {
        'flow': '"bst_691"'
    },

    'bst_691': {
        'vraag': '"bst_692"',

        'ja_flow': 'bst_691.mfbpjk==bst_691.mfbknr, bst_691.mfbpnr==bst_691.mfbpnr',
        'nee_flow': 'bst_691.mfbpnk==bst_691.mfbknr, bst_691.mfbpnr==bst_691.mfbpnr',

        'ja_actie': "'bst_691.mfbpja==bst_693.mfbanr'",
        'nee_actie': "'bst_691.mfbpna==bst_693.mfbanr'"
    },

    'bst_692': {
        'functie': '"bst_689"',
        'vraag_functie_parameter': "'bst_695'",
        'vraag_functie_waardelijst': "'bst_696'",
        'vraag_functie_attribuut': "'bst_697'",
    },

    'bst_693': {
        'teksten': ['bst_693.mfbanr==bst_921.txkode, bst_693.txmodu==bst_921.txmodu, %s==bst_921.txtsrt' % ZIEKENHUISTEKST],
        'bouwstenen': '"bst_694"'
    },

    'bst_695': {
        'vraag': '"bst_692"',
        'functie': '"bst_689"',
        'parameter': '"bst_685"'
    },

    'bst_696': {
        'vraag': '"bst_692"',
        'functie': '"bst_689"',
        'waardelijst': '"bst_699"'
    },

    'bst_697': {
        'vraag': '"bst_692"',
        'functie': '"bst_689"',
        'attribuut': '"bst_686"'
    }
}

aggregates = {
    'bst_693': {
        'tekst': ('txtext', 'self.teksten')
    }
}

proxies = {
    'bst_692': {
        'parameters': ('vraag_functie_parameter', 'parameter'),
        'waardelijsten': ('vraag_functie_waardelijst', 'waardelijst'),
        'attributen': ('vraag_functie_attribuut', 'attribuut')
    }
}


import os
import requests
from bs4 import BeautifulSoup

from files import file_numbers
from constants import SKIP_FIELDS, BESCHRIJVINGEN_FILE_EXTENSION

from config import BESCHRIJVINGEN_DIRECTORY, BESTAND_BESCHRIJVINGEN_URL_PREFIX, DATA_DIRECTORY


def get_bestand_filename(name, add_ext=False):
    filename = 'BST%sT' % name
    if add_ext:
        filename += BESCHRIJVINGEN_FILE_EXTENSION
    return filename


def get_bestand_htmlpath(name):
    return os.path.join(BESCHRIJVINGEN_DIRECTORY, get_bestand_filename(name, add_ext=True))


def get_bestand_url(file_no):
    return BESTAND_BESCHRIJVINGEN_URL_PREFIX + get_bestand_filename(file_no)


def get_inputpath(file_no):
    return os.path.join(DATA_DIRECTORY, get_bestand_filename(file_no, add_ext=False))


# https://techoverflow.net/2017/02/26/requests-download-file-if-it-doesnt-exist/
def download_file(filename, url):
    with open(filename, 'wb') as fout:
        response = requests.get(url, stream=True)
        response.raise_for_status()
        for block in response.iter_content(4096):
            fout.write(block)


def download_if_not_exists(filename, url):
    if not os.path.exists(filename):
        download_file(filename, url)
        return True
    return False


def refresh_htmls():
    for file_no in file_numbers:
        download_if_not_exists(get_htmlpath(file_no), get_bestand_url(file_no))


def get_bestand_html(file_no):
    html_input = get_bestand_htmlpath(file_no)
    html = BeautifulSoup(open(html_input).read(), features='html5lib')

    return html


def extract_struct(table):
    fields = []

    for entry in table.findChildren('tr')[1:]:

        _desc, sr, size, fmt, _pos = map(lambda x: x.text, entry.findChildren('td'))

        if sr:
            key = int(sr[0])
        else:
            key = 0

        if fmt == 'N':
            kind = 'Integer'
            if '+' in size or ',' in size:
                size = size.split('(')[0]
        else:
            kind = 'String'

        if entry.th.a:
            name = entry.th.a.text.lower()
        else:
            name = 'empty'

        if name in SKIP_FIELDS:
            size = int(size) * -1

        field = {
            'name': name,
            'size': int(size),
            'kind': kind,
            'key': key
        }

        fields.append(field)

    return fields


import struct
from parsing import get_bestand_html, get_inputpath, extract_struct
from db import session
from constants import SKIP_FIELDS
from files import file_numbers
from utils import load_class

priority_files = [] # ['902']


def priosort(l, prio):
    def prio_index(x):
        try:
            return prio.index(x)
        except ValueError:
            return len(prio)

    return sorted(l, key=lambda x: prio_index(x))


def populate_db():
    for file_no in priosort(file_numbers, priority_files):

        print('Parsing file %s ...' % get_inputpath(file_no))

        html = get_bestand_html(file_no)
        fields = extract_struct(html.body.find('table', attrs={'class': 'zindextable'}))

        fieldwidths = [x['size'] for x in fields]
        fmtstring = ' '.join('{}{}'.format(abs(fw), 'x' if fw < 0 else 's') for fw in fieldwidths)
        fieldstruct = struct.Struct(fmtstring)

        parse = lambda line: tuple(s.decode() for s in fieldstruct.unpack_from(line.encode()))

        data = []
        with open(get_inputpath(file_no), "r") as infile:
            for line in infile:
                # 0x1a / substitute character / eof
                if len(line) == 1:
                    break
                datum = parse(line)
                data.append(datum)

        table_name = 'bst_%s' % file_no

        column_tups = []

        for s in fields:

            name = s['name']
            if name in SKIP_FIELDS:
                continue

            kind = s['kind']
            # If this fails, we need to translate types
            assert kind in ['Integer', 'String']

            column_tups.append((name, kind))

        table_class = load_class('model.' + table_name)
        table_instance = table_class

        columns = [x[0] for x in column_tups]
        for d in data:
            t = dict(zip(columns, d))

            b = table_instance(**t)
            session.add(b)

        session.commit()


populate_db()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gstandaard",
    version="0.0.1",
    author="Mark Santcroos",
    author_email="m.a.santcroos@lumc.nl",
    description="Python package for working with the G-Standaard from Z-Index.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://git.lumc.nl/pongs/gstandaard-python",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from model import bst_685, bst_690, bst_691, bst_695
from wes import group_dict
from db import session


from traversal import trav_prot

all_fg_protocols = session.query(bst_690).\
            join(bst_691, bst_690.mfbpnr==bst_691.mfbpnr).\
            join(bst_695, bst_691.mfbvnr==bst_695.mfbvnr).\
            join(bst_685, bst_695.mfbpanr==bst_685.mfbpanr).\
            filter(bst_685.mfbpaoms.like('FG:%')).all()
print(len(all_fg_protocols))
print(len(set(all_fg_protocols)))


def tsitnr2mfbpanrs(session, tsitnr):
    result = session.query(bst_685).filter_by(mfbpitnr=tsitnr).all()
    return [x.mfbpanr for x in result]


results = {}
answers = {}
unknown = 0
known = 0
for sample in group_dict:    
    sample_result = 0
    for tsitnr in group_dict[sample]:
        
        if tsitnr in answers:
            tsitnr_result = answers[tsitnr]
        else:        
            tsitnr_result = 0

            try:
                sample_params = tsitnr2mfbpanrs(session, tsitnr)
            except:
                # No parameter found?
                unknown += 1
                print('unknown')
                continue

            for protocol in all_fg_protocols:
                score_teller = 0
                doorlopen_pad = []
                
                # Alleen voor testapotheken
                if protocol.mfbpwin == 'J':
                    continue
                    
                ret = trav_prot(protocol.flow, sample_params, score_teller, doorlopen_pad, debug=True)
                
                # print('Score teller: %d' % score_teller)
                print('Doorlopen pad:', doorlopen_pad)

                if ret:
                    tsitnr_result += 1
        
            answers[tsitnr] = tsitnr_result

        sample_result += tsitnr_result
        known += 1

    results[sample] = sample_result

assert len(results) == 1355
assert sum([results[x] > 0 for x in results]) == 1144



from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from model import Base
from config import database

engine = create_engine('sqlite:///%s' % database, echo=False)

Session = sessionmaker(bind=engine)
session = Session()
Base.metadata.create_all(engine)
session.commit()
